/* splash.c - Splash screen for graphical boot
 *
 * Copyright (C) 2003 Jonathan Blandford <jrb@redhat.com>
 * All rights reserved.
 *
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This file is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this; see the file COPYING.  If not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include "splash.h"

#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <sys/kd.h>
#include <sys/vt.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <X11/Xatom.h>
#include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>
#include <gdk-pixbuf/gdk-pixdata.h>
#include <vte/vte.h>
#include <libintl.h>
#include <config.h>
#include "gtkexpander.h"

#define _(x) dgettext (GETTEXT_PACKAGE, x)
#define N_(x) (x)


static const char *init_redirect[] = { "/sbin/change_console", "-f", NULL };
static GdkColor start_color = {0, 0xFFFF, 0xFFFF, 0xFFFF};
static GdkColor end_color = {0, 0x9090, 0x0d0d, 0x1212};
static GdkColor background = {0, 0xc6c6, 0x1212, 0x1919};
static GdkColor highlight = {0, 0x9090, 0x0d0d, 0x1212};
#define N_STEPS 20


static void splash_set_label_text (Splash *splash, const gchar  *label);


/*
 * We really want ot catch [<Alt>]<Ctrl>F1 to switch to vt1
 */
  
static int
is_a_console(int fd) {
    char arg;

    arg = 0;
    return(ioctl(fd, KDGKBTYPE, &arg) == 0);
}

static int
open_a_console(const char *path)
{
    int fd;

    fd = open(path, O_RDWR);
    if (fd < 0 && errno == EACCES)
        fd = open(path, O_WRONLY);
    if (fd < 0 && errno == EACCES)
        fd = open(path, O_RDONLY);
    if (fd < 0)
        return(-1);
    if (!is_a_console(fd)) {
        close(fd);
        return(-1);
    }
    return(fd);
}

static int
get_console_fd(void) {
    int fd;

    fd = open_a_console("/dev/tty");
    if (fd >= 0)
        return(fd);

    fd = open_a_console("/dev/tty0");
    if (fd >= 0)
        return(fd);

    fd = open_a_console("/dev/vc/0");
    if (fd >= 0)
        return(fd);

    fd = open_a_console("/dev/console");
    if (fd >= 0)
        return(fd);

    for (fd = 0; fd < 1024; fd++)
        if (is_a_console(fd))
            return(fd);

    return(-1);
}

void
switch_to_vt(int vt) 
{
    int ret, fd;

    fd = get_console_fd();
    if (fd < 0)
	    return;

    ret = ioctl(fd, VT_ACTIVATE, vt);
    close(fd);
}

int
get_active_vt (void)
{
    int fd;
    struct vt_stat vt_state;

    fd = get_console_fd();
    if (ioctl (fd, VT_GETSTATE, &vt_state) < 0)
        return -1;

   return vt_state.v_active;
}

static gboolean
key_pressed_event(GtkWidget *widget ATTRIBUTE_UNUSED,
                  GdkEventKey *event,
		  gpointer user_data ATTRIBUTE_UNUSED) {
    if ((event->keyval == GDK_F1) && (event->state & GDK_CONTROL_MASK)) {
        switch_to_vt(1);
        return(TRUE);
    }
    return(FALSE);
}

static void
set_bg_cursor (GdkCursorType cursor_type)
{
  GdkCursor *cursor;
  GdkDisplay *display;
  gint i, n_screens;

  display = gdk_display_get_default ();
  n_screens = gdk_display_get_n_screens (display);
  cursor = gdk_cursor_new (cursor_type);

  for (i = 0; i < n_screens; i++)
    {
      GdkScreen *screen;

      screen = gdk_display_get_screen (display, i);
      gdk_window_set_cursor (gdk_screen_get_root_window (screen), cursor);
    }
  gdk_cursor_unref (cursor);
}

static void
window_realize (GtkWindow *window,
		Splash *splash)
{
  GdkAtom atom;

  gtk_window_move (window, splash->x, splash->y);
  gtk_window_fullscreen (window);

  /*
   * Jonathan's hack ask the WM to move on the 2nd desktop to get it
   * out of the way only if metacity gets started by firstboot
   */
  atom = gdk_atom_intern ("_NET_WM_WINDOW_TYPE_DESKTOP", FALSE);
  gdk_property_change (GTK_WIDGET(window)->window,
  		       gdk_atom_intern ("_NET_WM_WINDOW_TYPE", FALSE),
  		       gdk_x11_xatom_to_atom (XA_ATOM), 32,
  		       GDK_PROP_MODE_REPLACE, (guchar *) &atom, 1);
}

static void
window_show (GtkWindow *window ATTRIBUTE_UNUSED,
	     Splash *splash ATTRIBUTE_UNUSED)
{
}

static void
event_box_realize (GtkWidget *event_box,
		   gpointer data ATTRIBUTE_UNUSED)
{
  gdk_window_set_background (event_box->window, &highlight);
}

static gboolean
throbber_timer (gpointer data)
{
  Splash *splash = data;
  if (++splash->throbber_step == splash->n_throbber_steps)
    splash->throbber_step = 0;

  gtk_image_set_from_pixbuf (GTK_IMAGE (splash->simple_throbber), splash->throbber_pixbuf[splash->throbber_step]);
  gtk_image_set_from_pixbuf (GTK_IMAGE (splash->details_throbber), splash->throbber_pixbuf[splash->throbber_step]);

  return TRUE;
}

static void
initialize_throbber (Splash *splash)
{
  GdkPixbuf *pixbuf;
  gint i, size;

#ifdef SPLASH_MAIN
  pixbuf = gdk_pixbuf_new_from_file ("throbber-anim.png", NULL);
#else
  pixbuf = gdk_pixbuf_new_from_file ("/usr/share/rhgb/throbber-anim.png", NULL);
#endif
  if (pixbuf == NULL) {
      fprintf(stderr, "Error: rhgb: %s\n", "Failed to locate pixbuf");
      exit(1);
  }
  size = gdk_pixbuf_get_height (pixbuf);
  splash->n_throbber_steps = gdk_pixbuf_get_width (pixbuf)/size;
  splash->throbber_pixbuf = g_malloc (sizeof (GdkPixbuf *) * splash->n_throbber_steps);

  for (i = 0; i < splash->n_throbber_steps; i++)
    {
      splash->throbber_pixbuf[i] = gdk_pixbuf_new_subpixbuf (pixbuf,
							     i*size, 0,
							     size, size);
    }
  splash->throbber_step = 1;
  gtk_image_set_from_pixbuf (GTK_IMAGE (splash->simple_throbber), splash->throbber_pixbuf[0]);
  gtk_image_set_from_pixbuf (GTK_IMAGE (splash->details_throbber), splash->throbber_pixbuf[0]);
  splash->throbber_timeout = g_timeout_add (83, throbber_timer, splash);
}

static void
switch_page (GtkWidget *widget,
	     gpointer   data)
{
  GtkExpander *expander = (GtkExpander*) widget;
  Splash *splash;
  gint page;

  splash = data;

  page = gtk_notebook_get_current_page (GTK_NOTEBOOK (splash->notebook));
  if (page == 0)
    splash_show_details (splash);
  else
    splash_hide_details (splash);
  gtk_expander_set_expanded (expander, ! gtk_expander_get_expanded (expander));
}

Splash *
splash_new (void)
{
  Splash *splash;
  gint screen_width;
  gint screen_height;
  GtkWidget *vbox, *vbox2, *align;
  GtkWidget *event_box;
  GtkWidget *table, *hbox;
  GtkWidget *expander;
  GdkScreen *screen;
  GdkRectangle rect;
  gchar *text;
  
  /* Set the busy cursor as quickly as we can */
  set_bg_cursor (GDK_WATCH);

  splash = g_new0 (Splash, 1);

  screen = gdk_screen_get_default ();
  gdk_screen_get_monitor_geometry (screen, 0, &rect);
  screen_width = rect.width;
  screen_height = rect.height;
  splash->x = rect.x;
  splash->y = rect.y;

  /* Make the window */
  splash->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_app_paintable (splash->window, TRUE);
  gtk_window_set_default_size (GTK_WINDOW (splash->window), screen_width, screen_height);
  gtk_widget_modify_bg (splash->window, GTK_STATE_NORMAL, &background);
  g_signal_connect (G_OBJECT (splash->window),
		    "realize",
		    (GCallback) window_realize,
		    splash);
  g_signal_connect (G_OBJECT (splash->window),
		    "show",
		    (GCallback) window_show,
		    splash);
#if 0
  g_signal_connect (G_OBJECT (splash->window),
		    "expose-event",
		    (GCallback) window_expose,
		    NULL);
#endif
  g_signal_connect (G_OBJECT (splash->window),
		    "key-press-event",
		    (GCallback) key_pressed_event,
		    NULL);
  gtk_widget_realize (splash->window);

  vbox = gtk_vbox_new (FALSE, 25);
  gtk_container_set_border_width (GTK_CONTAINER (vbox), 25);
  gtk_container_add (GTK_CONTAINER (splash->window), vbox);
  splash->notebook = gtk_notebook_new ();
  gtk_widget_modify_bg (splash->notebook, GTK_STATE_NORMAL, &highlight);

  gtk_notebook_set_show_tabs (GTK_NOTEBOOK (splash->notebook), FALSE);
  gtk_notebook_set_show_border (GTK_NOTEBOOK (splash->notebook), FALSE);
  gtk_box_pack_start (GTK_BOX (vbox), splash->notebook, TRUE, TRUE, 0);
  align = gtk_alignment_new (0.5, 0.5, 0.0, 0.0);
  gtk_notebook_append_page (GTK_NOTEBOOK (splash->notebook), align, NULL);

  /* Simple page */
  vbox2 = gtk_vbox_new (FALSE, 25);
  gtk_container_add (GTK_CONTAINER (align), vbox2);

  event_box = gtk_event_box_new ();
  gtk_widget_modify_bg (event_box, GTK_STATE_NORMAL, &highlight);
  g_signal_connect (G_OBJECT (event_box),
		    "realize",
		    (GCallback) event_box_realize,
		    NULL);
  g_signal_connect (G_OBJECT (event_box),
		    "key-press-event",
		    (GCallback) key_pressed_event,
		    NULL);
  gtk_box_pack_start (GTK_BOX (vbox2), event_box, FALSE, FALSE, 0);

  table = gtk_table_new (2, 3, FALSE);
  gtk_container_set_border_width (GTK_CONTAINER (table), 25);
  gtk_table_set_row_spacings (GTK_TABLE (table), 12);
  gtk_table_set_col_spacings (GTK_TABLE (table), 12);

  gtk_container_add (GTK_CONTAINER (event_box), table);
  splash->simple_progress = gtk_progress_bar_new ();
  gtk_widget_set_size_request (splash->simple_progress, 350, -1);
  gtk_table_attach (GTK_TABLE (table), splash->simple_progress, 0, 3, 0, 1, GTK_EXPAND | GTK_FILL, 0, 0, 0);

  splash->simple_label = gtk_label_new (NULL);
  gtk_misc_set_alignment (GTK_MISC (splash->simple_label), 1.0, 0.5);
  gtk_table_attach (GTK_TABLE (table), splash->simple_label, 1, 2, 1, 2, GTK_EXPAND | GTK_FILL, 0, 0, 0);

  splash->simple_throbber = gtk_image_new ();
  gtk_table_attach (GTK_TABLE (table), splash->simple_throbber, 2, 3, 1, 2, 0, 0, 0, 0);

  text = g_strdup_printf ("<span weight=\"light\" foreground=\"white\">%s</span>", _("Show _Details"));
  expander = gtk_expander_new (text);
  g_free (text);
  gtk_label_set_use_markup (GTK_LABEL (gtk_expander_get_label_widget (GTK_EXPANDER (expander))), TRUE);
  gtk_label_set_use_underline (GTK_LABEL (gtk_expander_get_label_widget (GTK_EXPANDER (expander))), TRUE);
  g_signal_connect (expander, "activate", (GCallback) switch_page, splash);
  align = gtk_alignment_new (0.5, 0.5, 0.0, 0.0);
  gtk_container_add (GTK_CONTAINER (align), expander);
  gtk_table_attach (GTK_TABLE (table), align, 0, 1, 1, 2, GTK_FILL, GTK_FILL, 5, 0);


  /* Details page */
  event_box = gtk_event_box_new ();
  g_signal_connect (G_OBJECT (event_box),
		    "realize",
		    (GCallback) event_box_realize,
		    NULL);  gtk_notebook_append_page (GTK_NOTEBOOK (splash->notebook), event_box, NULL);
  gtk_widget_modify_bg (event_box, GTK_STATE_NORMAL, &highlight);
  vbox2 = gtk_vbox_new (FALSE, 25);
  gtk_container_set_border_width (GTK_CONTAINER (vbox2), 25);
  gtk_container_add (GTK_CONTAINER (event_box), vbox2);
  hbox = gtk_hbox_new (FALSE, 25);
  gtk_box_pack_start (GTK_BOX (vbox2), hbox, FALSE, FALSE, 0);

  text = g_strdup_printf ("<span weight=\"light\" foreground=\"white\">%s</span>", _("Hide _Details"));
  expander = gtk_expander_new (text);
  g_free (text);
  gtk_expander_set_expanded (GTK_EXPANDER (expander), TRUE);
  gtk_label_set_use_markup (GTK_LABEL (gtk_expander_get_label_widget (GTK_EXPANDER (expander))), TRUE);
  gtk_label_set_use_underline (GTK_LABEL (gtk_expander_get_label_widget (GTK_EXPANDER (expander))), TRUE);
  g_signal_connect (expander, "activate", (GCallback) switch_page, splash);
  align = gtk_alignment_new (0.5, 0.5, 0.0, 0.0);
  gtk_container_add (GTK_CONTAINER (align), expander);
  gtk_box_pack_start (GTK_BOX (hbox), align, FALSE, FALSE, 0);

  splash->details_progress = gtk_progress_bar_new ();
  gtk_widget_set_size_request (splash->details_progress, 350, -1);
  align = gtk_alignment_new (0.5, 0.5, 0.0, 0.0);
  gtk_container_add (GTK_CONTAINER (align), splash->details_progress);
  gtk_box_pack_start (GTK_BOX (hbox), align, FALSE, FALSE, 0);

  splash->details_label = gtk_label_new (NULL);
  gtk_misc_set_alignment (GTK_MISC (splash->details_label), 1.0, 0.5);
  gtk_box_pack_start (GTK_BOX (hbox), splash->details_label, TRUE, TRUE, 0);
  splash->details_throbber = gtk_image_new ();
  gtk_box_pack_start (GTK_BOX (hbox), splash->details_throbber, FALSE, FALSE, 0);


  splash->details_terminal = vte_terminal_new();
  vte_terminal_set_font_from_string (VTE_TERMINAL (splash->details_terminal), "monospace 12");
  vte_terminal_fork_command (VTE_TERMINAL (splash->details_terminal),
			     "/sbin/change_console",
			     (char **) init_redirect,
			     NULL, "/",
			     FALSE, FALSE, FALSE);
  gtk_box_pack_start (GTK_BOX (vbox2), splash->details_terminal, TRUE, TRUE, 0);

  /* Distro logo */
    
  align = gtk_alignment_new (1.0, 0.5, 1.0, 1.0);
  gtk_box_pack_end (GTK_BOX (vbox), align, FALSE, FALSE, 0);

#ifdef SPLASH_MAIN
  splash->system_logo = gtk_image_new_from_file ("system-logo.png");
#else
  splash->system_logo = gtk_image_new_from_file ("/usr/share/rhgb/system-logo.png");
#endif
  gtk_misc_set_alignment (GTK_MISC (splash->system_logo), 1.0, 0.5);
  gtk_container_add (GTK_CONTAINER (align), splash->system_logo);
  
  initialize_throbber (splash);
  gtk_widget_show_all (splash->window);

  return splash;
}


static void
get_fade_level (gint  count,
		gint *r,
		gint *g,
		gint *b)
{
  guint c;
	   
  if (end_color.red >= start_color.red)
    {
      c = (end_color.red - ((end_color.red - start_color.red) * ((count - 1)/((float)N_STEPS))));
      *r = c / 256;
    }
  else
    {
      c = (start_color.red - ((start_color.red - end_color.red) * ((N_STEPS + 1 - count)/((float)N_STEPS))));
      *r = c / 256;
    }

  if (end_color.green >= start_color.green)
    {
      c = (end_color.green - ((end_color.green - start_color.green) * ((count - 1)/((float)N_STEPS))));
      *g = c / 256;
    }
  else
    {
      c = (start_color.green - ((start_color.green - end_color.green) * ((N_STEPS + 1 - count)/((float)N_STEPS))));
      *g = c / 256;
    }

  if (end_color.blue >= start_color.blue)
    {
      c = (end_color.blue - ((end_color.blue - start_color.blue) * ((count - 1)/((float)N_STEPS))));
      *b = c / 256;
    }
  else
    {
      c = (start_color.blue - ((start_color.blue - end_color.blue) * ((N_STEPS + 1 - count)/((float)N_STEPS))));
      *b = c / 256;
    }

}


static gboolean
label_fade_timeout (Splash *splash)
{
  if (splash->fade_step == 0)
    {
      splash_set_label_text (splash, "<span weight=\"light\"> </span>");
      splash->label_timeout = 0;
    }
  else
    {
      gchar *text;
      gint r, g, b;

      get_fade_level (splash->fade_step, &r, &g, &b);

      text = g_strdup_printf ("<span foreground=\"#%02X%02X%02X\" weight=\"light\">%s</span>", r, g, b, splash->label_text);
      splash_set_label_text (splash, text);
      g_free (text);
      splash->fade_step--;
      splash->label_timeout = g_timeout_add (1000/N_STEPS, (GSourceFunc) label_fade_timeout, splash);
    }
  return FALSE;
}


static gboolean
label_timeout_cb (Splash *splash)
{
  splash->fade_step = N_STEPS;
  splash->label_timeout = g_timeout_add (1000/N_STEPS, (GSourceFunc) label_fade_timeout, splash);
  return FALSE;
}


static void
setup_label_fade (Splash *splash)
{
  
  splash->label_timeout = g_timeout_add (5000, (GSourceFunc) label_timeout_cb, splash);
  splash->clean_label = FALSE;
}

static void
splash_set_label_text (Splash *splash,
		       const gchar  *label)
{
  gtk_label_set_markup (GTK_LABEL (splash->simple_label), label);
  gtk_label_set_markup (GTK_LABEL (splash->details_label), label);
}

void
splash_set_label (Splash *splash,
		  const gchar  *label)
{
  gchar *text;

  if (splash->label_timeout)
    {
      g_source_remove (splash->label_timeout);
      splash->label_timeout = 0;
    }

  if (splash->label_text)
    g_free (splash->label_text);
  splash->label_text = g_strdup (label ? label : " ");

  text = g_strdup_printf ("<span foreground=\"#%02X%02X%02X\" weight=\"light\">%s</span>",
			  start_color.red,
			  start_color.green,
			  start_color.blue,
			  splash->label_text);
  splash_set_label_text (splash, text);
  g_free (text);

  splash->clean_label = TRUE;
}

void
splash_set_progress (Splash  *splash,
		     gdouble  fraction)
{
  gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (splash->simple_progress), fraction);
  gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (splash->details_progress), fraction);
  
  if (splash->clean_label)
    setup_label_fade (splash);
}

void
splash_pulse (Splash *splash)
{
  gtk_progress_bar_pulse (GTK_PROGRESS_BAR (splash->simple_progress));
  gtk_progress_bar_pulse (GTK_PROGRESS_BAR (splash->details_progress));

  if (! splash->clean_label)
    setup_label_fade (splash);
}

void
splash_hide_screen (Splash *splash)
{
  gtk_widget_hide (splash->notebook);
  gtk_widget_hide (splash->system_logo);
}

void
splash_show_details (Splash *splash)
{
  gtk_notebook_set_current_page (GTK_NOTEBOOK (splash->notebook), 1);
  gtk_widget_grab_focus (splash->details_terminal);
}

void
splash_hide_details (Splash *splash)
{
  gtk_notebook_set_current_page (GTK_NOTEBOOK (splash->notebook), 0);
}

void
splash_echo (Splash  *splash, const gchar *string, glong length)
{
  const gchar *str_end, *cur;

  if (length <= 0)
      length = (glong) strlen(string);
  if (length <= 0)
      return;

  /* we need to convert \n to \n\r ... */
  str_end = &string[length];
  while ((string != NULL) && (string < str_end)) {
      cur = strchr(string, '\n');
      if (cur == NULL) {
	  vte_terminal_feed(VTE_TERMINAL(splash->details_terminal),
			    string, str_end - string);
	  break;
      } else {
	  vte_terminal_feed(VTE_TERMINAL(splash->details_terminal),
				  string, cur - string + 1);
	  vte_terminal_feed(VTE_TERMINAL(splash->details_terminal),
			    "\r", 1);
          string = cur + 1;
      }
  }
}

#ifdef SPLASH_MAIN
/*
 * Switching to vts
 */
int
main (int argc, char **argv)
{
  Splash *splash;
  gtk_init (&argc, &argv);

  bindtextdomain (GETTEXT_PACKAGE, "/usr/share/locale");
  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
  textdomain (GETTEXT_PACKAGE);

  splash = splash_new ();
  splash_set_progress (splash, 0.50);
  splash_set_label (splash, "Testing the rhgb splash screen");
  splash_echo(splash, "Testing vte output line 1\n", -1);
  splash_echo(splash, "Testing vte output line 2\nline 3\nunfinished line 4", -1);
 
  gtk_main ();

  return 0;
}
#endif
